
#ifndef __octree_hpp__
#define __octree_hpp__

#include <string>
#include <map>

#include "math3d.hpp"
#include "utils.hpp"
#include "threading.hpp"

class OctreeBuilder;

/* Octree tricubic interpolator. */
class Octree
{
public:
  Octree ();
  ~Octree ();

  operator bool () { return false; }

  bool getValues (const vect3d& pos, prec_t *scalars, vect3d *vectors);

  struct cell_t;
  struct subcells_t;

  struct interp_t
  {
    vect3d corner;
    float *fact;
  };

  struct subcells_t
  {
    vect3d middle;
    cell_t *sc[8];
  };

  struct cell_t
  {
    subcells_t *subcells;
    interp_t *interp;
    struct cell_t *next;
    size_t index;
  };

  bool save (const char *path);
  bool load (const char *path);

  friend class OctreeBuilder;

  void clear ();

  std::string getPovrayScript();

  struct level_t
  {
    vect3d bs; // box size
    vect3d hbs; // half box size
    vect3d rbs; // reciprocal box size
  };

  size_t size () const;

private:
  vect3d corner, box, hbox, rbox;
  size_t xc, yc, zc;
  typedef std::vector<level_t> levels_t;
  levels_t levels;

  size_t nScalars;
  size_t nVectors;
  size_t nValues;

  typedef std::vector<cell_t*> hash_t;
  hash_t hash;

  allocator<interp_t, 1024> a_interp;
  allocator<cell_t, 1024> a_cell;
  allocator<subcells_t, 1024> a_subcell;
  allocator<float, 4096> a_fact;

  typedef std::map<void *, size_t> cmap_t;

  void addPovrayScript(std::string& str, cell_t *cell, size_t level);
  void save (FPWriter& w, cmap_t& cmap, cell_t *cell);
  void get_stats (cell_t *cell, size_t& nCells, size_t& nInterp, size_t& nMirror, size_t& nSubcell,
      cmap_t& cmap);
};

/* Octree builder for tricubic interpolation. */
class OctreeBuilder
{
public:
  OctreeBuilder (const vect3d& c1 = vect3d(-1, -1, -1), const vect3d& c2 = vect3d(1, 1, 1),
      size_t nScalars = 1, size_t nVectors = 0, size_t limitMemoryMB = 100, prec_t smallestResolution = 0.0);
  virtual ~OctreeBuilder ();

  void addLimit (const vect3d& n);

  void build (Octree& tree);

  virtual void getValues (const vect3d& pos, prec_t *f, prec_t *dfdx, prec_t *dfdy, prec_t *dfdz,
      prec_t *d2fdxdy, prec_t *d2fdxdz, prec_t *d2fdydz, prec_t *d3fdxdydz) = 0;
  virtual bool inContact (const vect3d& pos) = 0;

private:
  vect3d c1, c2;
  size_t nScalars;
  size_t nVectors;
  size_t nValues;
  size_t limitMemoryMB;
  prec_t smallestResolution;

  struct values_str {
    values_str() : values(0) { }
    ~values_str () { delete [] values; }
    vect3d pos;
    prec_t *f, *dfdx, *dfdy, *dfdz, *d2fdxdy, *d2fdxdz, *d2fdydz, *d3fdxdydz;
    prec_t *values;
  };

  struct o8entry_t
  {
    Octree::cell_t *cell;
    vect3d c1, c2;
    values_str *vals[8];
  };

  typedef std::vector<vect3d> limits_t;
  limits_t limits;

  bool isInside(const vect3d& c1, const vect3d& c2);

  struct fieldtask : public Task
  {
    size_t part;
    OctreeBuilder *builder;
    std::vector<values_str*> *values_map_calc;
    void execute () { builder->run_calc_values (*values_map_calc, part); }
  };

  struct celltask : public Task
  {
    size_t part;
    OctreeBuilder *builder;
    std::vector<o8entry_t> *cells;
    vect3d box;
    void execute () { builder->run_calc_cells (*cells, part, box); }
  };

  void run_calc_values (std::vector<values_str*>& values_map_calc, size_t part);
  void run_calc_cells (std::vector<o8entry_t>& cells, size_t part, vect3d box);
};

#endif // !__octree_hpp__

