//
// Ephi - simulation of magnetic fields and particles
// Copyright (C) 2007 Indrek Mandre <indrek(at)mare.ee>
// For more information please see http://www.mare.ee/indrek/ephi/
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "ephi.hpp"
#include "potential.hpp"

#define LOOPS 10

static void usage (const char *name)
{
  fprintf (stderr, "%s: %s [-m pointsystem|linesystem|chargeflow] [-q] [-x] [-s] [-g] [-8] [-f] [-r resolution]\n", name, name);
  exit(-1);
}

int main (int argc, char *argv[])
{
  prec_t resolution = 0.018;
  bool quick = false;
  bool cross = false;
  bool gun = false;
  bool snapshots = false;
  bool at8 = false;
  bool fcage = false;
  PField::emethod_t method = PField::POINTSYSTEM;

  for ( int i = 1; i < argc; i++ )
    {
      if ( argv[i][0] != '-' )
          usage(argv[0]);
      switch (argv[i][1])
        {
          case 'm': {
            if ( ++i == argc )
                usage(argv[0]);
            if ( !strcmp (argv[i], "pointsystem") )
                method = PField::POINTSYSTEM;
            else if ( !strcmp (argv[i], "linesystem") )
                method = PField::LINESYSTEM;
            else if ( !strcmp (argv[i], "chargeflow") )
                method = PField::CHARGEFLOW;
            else
                usage(argv[0]);
            break;
          }
          case 'f': {
            fcage = true;
            break;
          }
          case '8': {
            at8 = true;
            break;
          }
          case 'g': {
            gun = true;
            break;
          }
          case 's': {
            snapshots = true;
            break;
          }
          case 'x': {
            cross = true;
            break;
          }
          case 'r': {
            if ( ++i == argc )
                usage(argv[0]);
            resolution = atof(argv[i]);
            break;
          }
          case 'q': {
            quick = true;
            break;
          }
          default: {
            usage(argv[0]);
            break;
          }
        }
    }

  prec_t cali_pot_min = 0;
  prec_t cali_pot_max = 0;
  prec_t cali_pot_add = 0;
  prec_t cali_e_min = 0;
  prec_t cali_e_max = 0;

  prec_t potential = 10000;
  prec_t r = 0.15;
  prec_t wr = 0.035;
  prec_t spacing = 0.01;
  prec_t rr = r + (2 * wr + spacing) / PREC_SQRT2;
  prec_t qinc = -1.5e-8;
  prec_t egun;
  vect3d egunpos;
  if ( at8 )
    {
      egunpos = vect3d(-1, 1, 1).normal() * (rr + r);
      egun = egunpos.y;
    }
  else
    {
      egunpos = vect3d(-(rr + r), 0, 0);
      egun = -egunpos.x;
    }

  FILE *csvfp = fopen ("polyp.csv", "w");
  if ( !csvfp )
    {
      fprintf (stderr, "%s: unable to open %s for writing\n", argv[0], "polyp.csv");
      exit(-1);
    }
  if ( gun )
      fprintf (csvfp, "Coil voltage[V],Space charge[C],Center potential[V],Potential well[V],Electron acceleration[V]\n");
  else
      fprintf (csvfp, "Coil voltage[V],Space charge[C],Coil charge[C],Center potential[V],Potential well[V],Electron acceleration[V]\n");

  for ( size_t i = LOOPS; i > 0; i-- )
    {
      char buf[32];
      sprintf (buf, "polyp%02u", (unsigned int)i);
      std::string prefix = buf;
      PField pf(method, resolution);
      prec_t q = (i - 1) * qinc;
#if 1
      pf.addTorus (vect3d(-rr, 0, 0), vect3d(1, 0, 0), r, wr, potential);
      pf.set_no_cdist (true);
      pf.addTorus (vect3d(rr, 0, 0), vect3d(-1, 0, 0), r, wr, potential);
      pf.set_no_cdist (false);
      pf.addTorus (vect3d(0, -rr, 0), vect3d(0, 1, 0), r, wr, potential);
      pf.addTorus (vect3d(0, rr, 0), vect3d(0, -1, 0), r, wr, potential);
      pf.addTorus (vect3d(0, 0, rr), vect3d(0, 0, -1), r, wr, potential);
      pf.addTorus (vect3d(0, 0, -rr), vect3d(0, 0, 1), r, wr, potential);

      if ( gun )
        {
          prec_t c = r / 4;
          if ( c < resolution )
              c = resolution;
          prec_t offset = egun + c + c / 4;

          if ( at8 )
            {
              pf.addSphere (vect3d(-offset, offset, offset), c, 0);
              pf.addSphere (vect3d(-offset, offset, -offset), c, 0);
              pf.addSphere (vect3d(-offset, -offset, offset), c, 0);
              pf.addSphere (vect3d(-offset, -offset, -offset), c, 0);
              pf.addSphere (vect3d(offset, -offset, offset), c, 0);
              pf.addSphere (vect3d(offset, -offset, -offset), c, 0);
              pf.addSphere (vect3d(offset, offset, offset), c, 0);
              pf.addSphere (vect3d(offset, offset, -offset), c, 0);
            }
          else
            {
              pf.addSphere (vect3d(-offset, 0, 0), c, 0);
              pf.addSphere (vect3d(offset, 0, 0), c, 0);
              pf.addSphere (vect3d(0, -offset, 0), c, 0);
              pf.addSphere (vect3d(0, offset, 0), c, 0);
              pf.addSphere (vect3d(0, 0, -offset), c, 0);
              pf.addSphere (vect3d(0, 0, offset), c, 0);
            }
        }

      if ( fcage )
        {
          pf.set_no_cdist (true);
          prec_t radius = egunpos.length() + r + r/2;
          prec_t ores = pf.setResolution (radius / 5.0);
          pf.addSphere (vect3d(0, 0, 0), radius, 0, true);
          pf.setResolution(ores);
          pf.set_no_cdist (false);
        }

      pf.setUniformCharge (vect3d(0, 0, 0), q, 0.1);

      if ( !pf.equalize() )
        {
          fprintf (stderr, "equalizing failed :(\n");
          abort();
        }
      pf.save (prefix + "_model.pot");
#else
      pf.load (prefix + "_model.pot");
#endif

      prec_t gunaccel = 0;
      vect3d spos = egunpos;
      vect3d gdir = -egunpos / 1000.0;
      for ( size_t k = 0; k < 1000; k++ )
        {
          prec_t pot = pf.getPotential (spos);
          if ( pot < gunaccel )
              break;
          gunaccel = pot;
          spos += gdir;
        }

      vect3d pot_pos(0, 0, 0);
      prec_t pot_well = pf.getPotential (pot_pos);
      while (1)
        {
          pot_pos = pot_pos + (r / 1000.0) * vect3d(1, 0, 0);
          prec_t pot = pf.getPotential (pot_pos);
          if ( pot < pot_well ) break;
          pot_well = pot;
        }

      std::string str = pf.getChargeConcentrationScript ();
      FILE *fp = fopen ((prefix + "_charge.pov").c_str(), "w");
      fprintf (fp, "camera { location <3, 3, 2>*0.25 look_at <0, 0, 0> sky <0, 0, 1> right x*image_width/image_height }\n");
      fprintf (fp, "#include \"scene.inc\"\n");
      fprintf (fp, "%s", str.c_str());
      fclose (fp);

      str = pf.getPovrayScript ();
      fp = fopen ((prefix + "_polywell.pov").c_str(), "w");
      fprintf (fp, "camera { location <3, 3, 2>*0.25 look_at <0, 0, 0> sky <0, 0, 1> right x*image_width/image_height }\n");
      fprintf (fp, "#include \"scene.inc\"\n");
      fprintf (fp, "%s", str.c_str());
      fclose (fp);

      fp = fopen ((prefix + "_info.txt").c_str(), "w");
      prec_t added_charge = pf.getAddedCharge();
      fprintf (fp, "Coil voltage = %.2f V\n", prec2double(potential));
      if ( !gun )
        {
          fprintf (fp, "Coil charge = %.5e C\n", prec2double(added_charge));
          fprintf (fp, "Coil Capacitance = %.3f pF\n", prec2double(added_charge / potential)*1e12);
        }
      fprintf (fp, "Space charge = %.5e C\n", prec2double(q));
      prec_t pot_mid = pf.getPotential(vect3d(0, 0, 0));
      fprintf (fp, "Potential in the middle: %.2f V\n", prec2double(pot_mid));
      fprintf (fp, "Potential well: %.2f V\n", prec2double(fabs(pot_mid - pot_well)));
      prec_t atgun;
      if ( gun )
          atgun = 0;
      else
          atgun = pf.getPotential (egunpos);
      prec_t eacc = gunaccel - atgun;
      fprintf (fp, "Electron acceleration: %.2f V\n", prec2double(eacc));
      fclose (fp);

      if ( !gun )
        {
          fprintf (csvfp, "%g,%g,%g,%g,%g,%g\n",
              prec2double (potential), prec2double(q), prec2double(added_charge),
              prec2double (pot_mid), prec2double(pot_well - pot_mid), prec2double(eacc));
        }
      else
        {
          fprintf (csvfp, "%g,%g,%g,%g,%g\n", prec2double (potential), prec2double(q),
              prec2double (pot_mid), prec2double(pot_well - pot_mid), prec2double(eacc));
        }
      fflush (csvfp);

      if ( quick )
          continue;

      Screen screen(800, 800, 1.4, 1.4);
      if ( cross )
          screen.set_axis (vect3d(0, 0, 0), vect3d(1, 1, 0).normal(), vect3d(0, 0, 1));
      Statics statics;
      ElectroDynamics dyn(statics);
      statics.addStaticElement (new StaticPotentialField (pf));
      Scene scene(dyn);

      scene.calc(screen, Scene::CALC_POTENTIAL);
      scene.set_coloring (20, false);
      if ( i == LOOPS )
        {
          cali_pot_min = scene.get_fmin();
          cali_pot_max = scene.get_fmax();
          cali_pot_add = scene.get_fadd(); // potential can be negative
        }
      else
        {
          scene.calibrate_field (cali_pot_min, cali_pot_max, cali_pot_add);
        }
      scene.render_map (screen);
      screen.write (prefix + "_potential.bmp");

      scene.calc(screen, Scene::CALC_EFIELD);
      scene.set_coloring (0, false);
      if ( i == LOOPS )
        {
          cali_e_min = scene.get_fmin();
          cali_e_max = scene.get_fmax();
        }
      else
        {
          scene.calibrate_field (cali_e_min, cali_e_max);
        }
      scene.render_map (screen);
      screen.write (prefix + "_efield.bmp");
      scene.set_coloring (20, true);
      scene.render_map (screen);
      screen.write (prefix + "_efield_log_stepped.bmp");
      scene.set_coloring (0, true);
      scene.render_lic (screen);
      screen.write (prefix + "_efield_lic_logcolored.bmp");
      if ( snapshots )
        {
          Screen screen(380, 380, 1.4, 1.4);
          if ( cross )
              screen.set_axis (vect3d(0, 0, 0), vect3d(1, 1, 0).normal(), vect3d(0, 0, 1));
          Statics statics;
          ElectroDynamics dyn(statics);
          statics.addStaticElement (new StaticPotentialField (pf));
          Scene scene(dyn);

          scene.calc(screen, Scene::CALC_POTENTIAL);
          scene.set_coloring (20, false);
          scene.calibrate_field (cali_pot_min, cali_pot_max, cali_pot_add);
          scene.render_map (screen);
          screen.write (prefix + "_min_potential.bmp");

          scene.calc(screen, Scene::CALC_EFIELD);
          scene.set_coloring (0, false);
          scene.calibrate_field (cali_e_min, cali_e_max);
          scene.render_map (screen);
          screen.write (prefix + "_min_efield.bmp");
          scene.set_coloring (20, true);
          scene.render_map (screen);
          screen.write (prefix + "_min_efield_log_stepped.bmp");
          scene.set_coloring (0, true);
          scene.render_lic (screen);
          screen.write (prefix + "_min_efield_lic_logcolored.bmp");
        }
    }

  fclose (csvfp);

  return 0;
}

