//
// Ephi - simulation of magnetic fields and particles
// Copyright (C) 2007 Indrek Mandre <indrek(at)mare.ee>
// For more information please see http://www.mare.ee/indrek/ephi/
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

// Polywell simulation. Simulate the movement of a single electron
// inside the polywell magnetic containment system.

#ifndef __polywell_conf_hpp__
#define __polywell_conf_hpp__

/** Abstract class for coil generation. */
class CoilFactory
{
public:
  virtual ~CoilFactory () { }

  /** Make a coil. */
  virtual bool make (Statics& statics, const vect3d& pos, const vect3d& normal,
      prec_t radius, prec_t wr, prec_t I, prec_t q_per_m) const = 0;

  /** Make an opposing coilpair with specified distance apart. */
  virtual bool make_coilpair (Statics& statics, const vect3d& pos, const vect3d& normal,
      prec_t distance, prec_t radius, prec_t wr, prec_t I, prec_t q_per_m) const
  {
    vect3d n = normal.normal();
    return make (statics, pos + 0.5 * distance * n, -normal, radius, wr, I, q_per_m) &&
        make (statics, pos - 0.5 * distance * n, normal, radius, wr, I, q_per_m);
  }
};

/** Generate coils using a specific shape matrix. */
class ShapedCoilFactory : public CoilFactory
{
public:
  ShapedCoilFactory (int n = 1);
  ShapedCoilFactory (int w, int h, prec_t hs = 1.0, prec_t vs = 1.0);
  virtual ~ShapedCoilFactory();

  bool make (Statics& statics, const vect3d& pos, const vect3d& normal,
      prec_t radius, prec_t wr, prec_t I, prec_t q_per_m) const;

  void set(int x, int y, bool v);
  void set_all (bool v);
  void set_circle ();

  int get_width () const { return width; }
  int get_height () const { return height; }
  prec_t get_hscale () const { return hscale; }
  prec_t get_vscale () const { return vscale; }
  void set_scale (prec_t hs, prec_t vs) { hscale = hs; vscale = vs; }

  bool get(int x, int y, vect3d& offset) const;
  int count() const;
  void print ();

private:
  int width, height;
  prec_t hscale, vscale;
  std::vector<bool> grid;
};

/** Create a dodecahedron polywell. */
extern void make_polywell_dodecahedron (Statics& statics, prec_t R, prec_t magnet_radius,
    prec_t magnet_spacing, prec_t I, prec_t q_per_m = 0, const CoilFactory& cf = ShapedCoilFactory());
/** Create a cube polywell. */
extern void make_polywell_cube (Statics& statics, prec_t R, prec_t magnet_radius,
    prec_t magnet_spacing, prec_t I, prec_t q_per_m = 0, const CoilFactory& cf = ShapedCoilFactory());
/** Create a solenoid. */
extern void make_solenoid (Statics& statics, size_t n, prec_t R, prec_t wire_radius,
    prec_t spacing, prec_t I, const CoilFactory& cf = ShapedCoilFactory());

extern const transf3d_pair& cube_polywell_transform (vect3d& p);

#endif // !__polywell_conf_hpp__

