//
// Ephi - simulation of magnetic fields and particles
// Copyright (C) 2007 Indrek Mandre <indrek(at)mare.ee>
// For more information please see http://www.mare.ee/indrek/
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include <math.h>
#include <assert.h>

#include "ephi.hpp"
#include "potential.hpp"

#include <set>

//#define PDIFF 0.009
//#define PDIFF 0.0115 // for polywell 15cm radius
//#define PDIFF 0.0090 // for 0.25m radius sphere
//#define PDIFF 0.009 // for 0.25m-0.075 radius torus
//#define PDIFF 0.018 // for polyp runs
//#define PDIFF 0.032 // for testing

#define PCONST (SPEED_OF_LIGHT * SPEED_OF_LIGHT * MAGNETIC_CONSTANT)

PField::PField (emethod_t method, prec_t resolution) : method(method), resolution(resolution)
{
  pc_q = 0;
  pc_pos.clear();
  pcounter = 0;
  added_charge = 0;
  cache = 0;
  no_cdist = false;
}

PField::~PField ()
{
}

void PField::setPointCharge (const vect3d& pos, prec_t q)
{
  setUniformCharge (pos, q, 0.0);
}

void PField::setUniformCharge (const vect3d& pos, prec_t q, prec_t radius)
{
  pc_q = q;
  pc_pos = pos;
  pc_radius = radius;
  pc_rr = radius * radius;
  if ( radius != 0.0 )
      pc_rho = 3 * q / (4.0 * PREC_PI * radius * radius * radius);
}

PPoint& PField::addPoint (const vect3d& pos, const vect3d& ppos, prec_t potential)
{
  points.push_back (PPoint(pos, ppos, potential));
  points.back().id = pcounter++;
  if ( no_cdist )
      points.back().flags |= PPoint::NO_CDIST;
  return points.back();
}

void PField::addFeedPoint (PPoint& p, prec_t potential)
{
  if ( potential == UNKNOWN_POTENTIAL )
      return;
  fpoint_t fp;
  fp.point = &p;
  fp.potential = potential;
  feeds.push_back (fp);
}

struct pair_t {
  pair_t (PPoint *p, prec_t phi) : p(p), phi(phi) { }
  PPoint *p;
  prec_t phi;
};

void PField::addSphere (const vect3d& pos, prec_t radius, prec_t potential, bool backwards)
{
#if 1
  // my geometric shapes knowldge really sux
  // the idea is that we make a icosahedron here and split it until we get to desired resolution
  prec_t phi = (1+prec_t_sqrt(5.0))*0.5; // golden rectangle ratio
  prec_t elen = 2 * radius / prec_t_sqrt (phi * prec_t_sqrt(5.0));

  prec_t potoff = resolution * 3;
  if ( potoff > radius * 0.2 )
      potoff = radius * 0.3;
  prec_t pratio = potoff/radius;
  if ( backwards )
      pratio = -pratio;

  std::vector<PPoint*> spoints;
  vect3d p;
  p = pos + vect3d(0, -1, -phi).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(0, -1, phi).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(0, 1, -phi).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(0, 1, phi).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));

  p = pos + vect3d(-1, -phi, 0).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(-1, phi, 0).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(1, -phi, 0).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(1, phi, 0).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));

  p = pos + vect3d(-phi, 0, -1).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(-phi, 0, 1).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(phi, 0, -1).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));
  p = pos + vect3d(phi, 0, 1).normal() * radius; spoints.push_back (&addPoint(p, p + (pos - p) * pratio, potential));

  std::vector<PLink*> slinks;
  slinks.reserve(30);
  for ( size_t i = 0; i < spoints.size(); i++ )
    {
      addFeedPoint (*spoints[i], potential);
      for ( size_t j = 0; j < spoints.size(); j++ )
        {
          if ( 0.9 * (spoints[i]->pos - spoints[j]->pos).length() < elen )
            {
              PLink *l = safe_connect (*spoints[i], *spoints[j]);
              if ( l ) slinks.push_back (l);
            }
        }
    }

  // now we need to start splitting the icosahedron more liking to our size
  while ( elen > resolution )
    {
      // first create mid-points
      for ( size_t i = 0; i < slinks.size(); i++ )
        {
          PLink& l = *slinks[i];
          vect3d p = 0.5 * (l.a->pos + l.b->pos);
          p = pos + (p - pos).normal() * radius;
          l.tmp = &addPoint (p, p + (pos - p) * pratio, potential);
        }

      // find connections between midpoints
      std::vector<std::pair<PPoint*,PPoint*> > tc;
      tc.reserve (slinks.size() * 2);
      for ( size_t i = 0; i < spoints.size(); i++ )
        {
          PPoint *a = spoints[i];
          for ( PPoint::links_t::iterator alit = a->links.begin(); alit != a->links.end(); alit++ )
            {
              PLink *l1 = *alit;
              PPoint *b = l1->a == a ? l1->b : l1->a;
              for ( PPoint::links_t::iterator blit = b->links.begin(); blit != b->links.end(); blit++ )
                {
                  PLink *l2 = *blit;
                  if ( l1 == l2 ) continue;
                  PPoint *c = l2->a == b ? l2->b : l2->a;
                  for ( PPoint::links_t::iterator clit = c->links.begin(); clit != c->links.end(); clit++ )
                    {
                      PLink *l3 = *clit;
                      if ( l2 == l3 ) continue;
                      PPoint *d = l3->a == c ? l3->b : l3->a;
                      if ( d != a ) continue;
                      tc.push_back (std::make_pair(l1->tmp, l2->tmp));
                      tc.push_back (std::make_pair(l2->tmp, l3->tmp));
                      tc.push_back (std::make_pair(l3->tmp, l1->tmp));
                      // we have found a triangle now, formed between l1, l2 and l3
                    }
                }
            }
        }

      // connect midpoints
      std::vector<PLink*> new_slinks;
      new_slinks.reserve (slinks.size() * 4);
      for ( size_t i = 0; i < tc.size(); i++ )
        {
          PLink *l = safe_connect (*tc[i].first, *tc[i].second);
          if ( l ) new_slinks.push_back (l);
        }

      // finally split up old links
      spoints.reserve (spoints.size() * 4);
      for ( size_t i = 0; i < slinks.size(); i++ )
        {
          PLink&l = *slinks[i];
          spoints.push_back (l.tmp);
          new_slinks.push_back (&l);
          new_slinks.push_back (&split(l, *l.tmp));
        }
      slinks.swap (new_slinks);
      elen = elen * 0.5;
    }

  if (0)
  for ( size_t i = 0; i < spoints.size(); i++ )
      printf ("%.32g\n", prec2double ((spoints[i]->pos - pos).length()));

#else
  vect3d normal(1, 0, 0);

  prec_t angle = asin(prec2double(resolution/radius));

  size_t n = (size_t)ceil(PREC_PI / angle);
  if ( (n % 2) ) n++;
  prec_t a = PREC_PI / n;

  std::vector<pair_t> last_row;
  std::vector<pair_t> next_row;

  prec_t potoff = resolution * 3;
  if ( potoff > radius * 0.2 )
      potoff = radius * 0.3;

  for ( size_t i = 0; i <= n; i++ )
    {
      prec_t offset = radius * cos(a * i);
      vect3d c = pos - offset * normal;
      prec_t sr = prec_t_sqrt(radius * radius - offset * offset);

      prec_t circum = sr * 2 * PREC_PI;
      size_t l = (size_t)ceil(circum / resolution);
      if ( !l ) l = 1;
      prec_t aadd = 2 * PREC_PI / l;

      next_row.clear();
      next_row.reserve (l);

      prec_t arand = (i % 2) ? aadd * 0.5 : 0;
      arand = 0;

      for ( size_t j = 0; j < l; j++ )
        {
          prec_t y, z;
          prec_t phi = j * aadd + arand;
          y = cos(phi) * sr;
          z = sin(phi) * sr;

          vect3d pl = c + vect3d(0, y, z);
          PPoint* point = &addPoint (pl, pl - (pl - pos).normal() * potoff, potential);
          if ( i == 0 || i == n || n > 5 && i == n / 2 && !(j % 4) && j + 2 < l )
              addFeedPoint (*point, potential);
          if ( !next_row.empty() )
              connect (*point, *next_row.back().p);
          if ( j > 1 && j == l - 1 )
              connect (*point, *next_row[0].p);
          next_row.push_back (pair_t(point, phi));
        }

#if 1
      bool swap = true;
      //printf ("row\n");
      if ( !last_row.empty() )
        {
          if ( last_row.size() > next_row.size() )
            {
              last_row.swap (next_row);
              swap = false;
            }

          if ( last_row.size() != 1 )
            {
              //printf ("preconnect 0(%f) ->%zu(%f)\n", next_row[0].phi, last_row.size()-1, last_row[last_row.size()-1].phi);
              connect (*next_row[0].p, *last_row[last_row.size()-1].p);
            }
          //printf ("%u vs %u\n", next_row.size(), last_row.size());
          size_t k = 0;
          for ( size_t j = 0; j < next_row.size(); j++ )
            {
              //printf ("connect %zu(%f) ->%zu(%f)\n", j, next_row[j].phi, k, last_row[k].phi);
              connect (*next_row[j].p, *last_row[k].p);
              while ( next_row[j].phi > last_row[k].phi && last_row.size() != 1 )
                {
                  k++;
                  if ( k >= last_row.size() )
                      break;
                  //printf ("subconnect %zu(%f) ->%zu(%f)\n", j, next_row[j].phi, k, last_row[k].phi);
                  connect (*next_row[j].p, *last_row[k].p);
                }
            }
        }

      if ( swap )
#endif
          last_row.swap (next_row);
    }
#endif
}

void PField::addTorus (const vect3d& pos, const vect3d& n, prec_t radius, prec_t wr, prec_t potential)
{
  vect3d normal = n.normal();
  vect3d r;
  normal.across(r);
  r = r * radius;
  vect3d p0 = pos + r;
  vect3d point_r = normal * wr;

  prec_t potway = resolution * 3;
  if ( potway > 0.75 * wr )
      potway = 0.75 * wr;

  size_t segment_count = (size_t)ceil((radius + wr) * 2 * PREC_PI / (2 * resolution / 2));
  while ( segment_count%4 ) segment_count++;
  size_t point_count = (size_t)ceil(wr * 2 * PREC_PI / resolution);
  while ( point_count%4 ) point_count++;

  if ( method == LINESYSTEM && 0 )
    {
      segment_count /= 4;
      //point_count /= 2;
    }

  prec_t segment_angle = 2 * PREC_PI / segment_count;
  prec_t point_angle = 2 * PREC_PI / point_count;

  vect3d p0_pc = p0 - pos;
  vect3d nXp0_pc = normal % p0_pc;

  std::vector<PPoint*> first_row;
  std::vector<PPoint*> last_row;

  for ( size_t i = 0; i < segment_count; i++ )
    {
      prec_t phi = segment_angle * i;
      vect3d center = pos + cos(phi) * p0_pc + sin(phi) * nXp0_pc;
      vect3d slope = -sin(phi) * p0_pc + cos(phi) * nXp0_pc;

      {
        vect3d pos = center;
        vect3d normal = slope.normal();
        vect3d p0 = pos + point_r;
        vect3d p0_pc = p0 - pos;
        vect3d nXp0_pc = normal % p0_pc;

        std::vector<PPoint*> next_row;
        for ( size_t j = 0; j < point_count; j++ )
          {
            prec_t phi = point_angle * j;
            vect3d mid = pos + cos(phi) * p0_pc + sin(phi) * nXp0_pc;
            vect3d pot_mid = mid - (mid - pos).normal() * potway;
            PPoint &p = addPoint (mid, pot_mid, potential);
            if ( (method != LINESYSTEM || 0) && !next_row.empty() )
                connect (*next_row.back(), p);
            next_row.push_back (&p);
            if ( !last_row.empty() )
                connect (*last_row[j], *next_row[j]);
            if ( i == segment_count - 1 )
                connect (*next_row[j], *first_row[j]);

            if ( j == point_count / 4 && i && !(i % 4) )
                addFeedPoint(p, potential);
          }
        if ( method != LINESYSTEM || 0 )
            connect (*next_row.back(), *next_row[0]);

        if ( i == 0 )
            first_row = next_row;
        last_row.swap (next_row);
      }
    }
}

void PField::addStrip (const vect3d& p0, const vect3d& p1, size_t n)
{
  prec_t len = (p1 - p0).length();
  if ( !n )
      n = (size_t)ceil(len/resolution);
  prec_t hf = len / n;
  vect3d p01 = (p1 - p0).normal();

  PPoint *prev = 0;
  for ( size_t i = 0; i <= n; i++ )
    {
      PPoint *p = &addPoint (p0 + i * hf * p01);
      if ( prev )
          connect (*prev, *p);
      prev = p;
    }
}

void PField::addT (const vect3d& pa, const vect3d& pb, const vect3d& pc, const vect3d& pd)
{
  PPoint& a = addPoint (pa);
  PPoint& b = addPoint (pb);
  PPoint& c = addPoint (pc);
  PPoint& d = addPoint (pd);
  connect (a, b);
  connect (b, c);
  connect (b, d);
}

void PField::print ()
{
#if 1
  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      char buf2[128];
      it->pos.sprint (buf2);
      printf ("%3u: %s: q=%g, rho=%g, pot=%g\n", (unsigned int)it->id, buf2, prec2double(it->charge),
          prec2double(it->rho), prec2double (getPotential(it->pot_pos)));
    }
#endif

  prec_t totpot = 0, minpot = 9e99, maxpot = -9e99;
  for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
    {
#if 0
      it->mid.print("link.mid");
#endif
      prec_t pot = getPotential(it->mid);
      if ( minpot > pot )
          minpot = pot;
      if ( maxpot < pot )
          maxpot = pot;
#if 0
      printf("pot = %g\n", getPotential(it->mid));
#endif
      totpot += pot;
    }
  printf ("pot.avg=%g, pot.min=%g, pot.max=%g\n", prec2double(totpot / links.size()), prec2double(minpot), prec2double(maxpot));
  printf ("Total points: %u, total links: %u\n", (unsigned int)points.size(), (unsigned int)links.size()); 
}

prec_t PField::getAveragePotential ()
{
  prec_t potadd = 0;
  for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
      potadd += getPotential(it->mid);
  return potadd / links.size();
}

void PField::setPotential (prec_t potential)
{
  prec_t ratio = potential / getAveragePotential();
  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
      it->charge *= ratio;
  added_charge *= ratio;
}

namespace {
  const char *PFMAGIC = "pFm1";

  struct pfheader
  {
    char magic[4];
    unsigned int pcounter;
    unsigned int points;
    unsigned int links;
    unsigned int feeds;
    double added_charge;
    double pc_q;
    double pc_pos_x, pc_pos_y, pc_pos_z;
    double pc_radius;
  };

  struct pfpoint
  {
    unsigned int id;
    double pos_x, pos_y, pos_z;
    double pot_pos_x, pot_pos_y, pot_pos_z;
    double charge, rho;
    size_t flags;
  };

  struct pflink
  {
    unsigned int id1;
    unsigned int id2;
  };

  struct pffeed
  {
    unsigned int id;
    prec_t potential;
  };
}

void PField::save (const char *path)
{
  FILE *fp = fopen (path, "wb");
  if ( !fp )
    {
      Ephi::debug (Ephi::DEBUG_CRITICAL, "PField::save: unable to open %s for writing!\n", path);
      return;
    }

  pfheader h;
  memcpy (h.magic, PFMAGIC, 4);
  h.pcounter = pcounter;
  h.points = points.size();
  h.links = links.size();
  h.feeds = feeds.size();
  h.added_charge = prec2double (added_charge);
  h.pc_q = prec2double (pc_q);
  h.pc_pos_x = prec2double (pc_pos.x);
  h.pc_pos_y = prec2double (pc_pos.y);
  h.pc_pos_z = prec2double(pc_pos.z);
  h.pc_radius = prec2double (pc_radius);
  fwrite (&h, sizeof(h), 1, fp);

  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      pfpoint pfp;
      pfp.id = it->id;
      pfp.pos_x = prec2double(it->pos.x);
      pfp.pos_y = prec2double(it->pos.y);
      pfp.pos_z = prec2double(it->pos.z);
      pfp.pot_pos_x = prec2double(it->pot_pos.x);
      pfp.pot_pos_y = prec2double(it->pot_pos.y);
      pfp.pot_pos_z = prec2double(it->pot_pos.z);
      pfp.charge = prec2double(it->charge);
      pfp.rho = prec2double(it->rho);
      pfp.flags = it->flags;
      fwrite (&pfp, sizeof(pfp), 1, fp);
    }

  for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
    {
      pflink pfl;
      pfl.id1 = it->a->id;
      pfl.id2 = it->b->id;
      fwrite (&pfl, sizeof(pfl), 1, fp);
    }

  for ( feeds_t::iterator it = feeds.begin(); it != feeds.end(); it++ )
    {
      pffeed pff;
      pff.id = it->point->id;
      pff.potential = prec2double (it->potential);
      fwrite (&pff, sizeof(pff), 1, fp);
    }

  fclose (fp);
}

bool PField::load (const char *path)
{
  links.clear();
  feeds.clear();
  points.clear();

  FILE *fp = fopen (path, "rb");
  if ( !fp )
      return false;

  pfheader h;
  if ( !fread (&h, sizeof(h), 1, fp) )
    {
      Ephi::debug (Ephi::DEBUG_ERROR, "PField::load: unable to load the header\n");
      fclose (fp);
      return false;
    }

  if ( memcmp (h.magic, PFMAGIC, 4) )
    {
      Ephi::debug (Ephi::DEBUG_ERROR, "PField::load: invalid magic\n");
      fclose (fp);
      return false;
    }

  //h.points = points.size();
  //h.links = links.size();
  //h.feeds = feeds.size();

  pcounter = h.pcounter;
  added_charge = h.added_charge;
  setUniformCharge (vect3d(h.pc_pos_x, h.pc_pos_y, h.pc_pos_z), h.pc_q, h.pc_radius);

  std::map<size_t, PPoint*> phash;

  for ( size_t i = 0; i < h.points; i++ )
    {
      pfpoint pfp;
      if ( !fread (&pfp, sizeof(pfp), 1, fp) )
        {
          Ephi::debug (Ephi::DEBUG_ERROR, "PField::load: unable to load the point %u\n", (unsigned int)i);
          fclose (fp);
          return false;
        }
      vect3d pos(pfp.pos_x, pfp.pos_y, pfp.pos_z);
      vect3d pot_pos(pfp.pot_pos_x, pfp.pot_pos_y, pfp.pot_pos_z);
      points.push_back (PPoint(pos, pot_pos));
      points.back().id = pfp.id;
      points.back().charge = pfp.charge;
      points.back().rho = pfp.rho;
      points.back().flags = pfp.flags;
      phash[pfp.id] = &points.back();
    }

  for ( size_t i = 0; i < h.links; i++ )
    {
      pflink pfl;
      if ( !fread (&pfl, sizeof(pfl), 1, fp) )
        {
          Ephi::debug (Ephi::DEBUG_ERROR, "PField::load: unable to load the link %u\n", (unsigned int)i);
          fclose (fp);
          return false;
        }
      std::map<size_t,PPoint*>::iterator p1it = phash.find(pfl.id1);
      std::map<size_t,PPoint*>::iterator p2it = phash.find(pfl.id2);
      if ( p1it == phash.end() || p2it == phash.end() )
        {
          Ephi::debug (Ephi::DEBUG_ERROR, "PField::load: unable to find link points at %u\n", (unsigned int)i);
          fclose (fp);
          return false;
        }
      connect(*p1it->second, *p2it->second);
    }

  for ( size_t i = 0; i < h.feeds; i++ )
    {
      pffeed pff;
      if ( !fread (&pff, sizeof(pff), 1, fp) )
        {
          Ephi::debug (Ephi::DEBUG_ERROR, "PField::load: unable to load feed %u\n", (unsigned int)i);
          fclose (fp);
          return false;
        }
      std::map<size_t,PPoint*>::iterator fit = phash.find(pff.id);
      if ( fit == phash.end() )
        {
          fclose (fp);
          return false;
        }
      addFeedPoint (*fit->second, pff.potential);
    }

  fclose (fp);

  return true;
}

#define PRINTMOD 10000

void PField::calc_efield (const links_t::iterator& begin, const links_t::iterator& end, const vect3d* cache)
{
  if ( cache )
    {
      for ( links_t::iterator it = begin; it != end; it++ )
        {
          PLink& c = *it;
          vect3d efield;
          efield.clear();
          for ( points_t::iterator jt = points.begin(); jt != points.end(); jt++ )
              efield += jt->charge * *cache++;

          if ( pc_q != 0 )
              efield += pc_q * *cache++;

          c.q = efield * c.cc_normal;
        }
    }
  else
    {
      for ( links_t::iterator it = begin; it != end; it++ )
        {
          PLink& c = *it;
          vect3d efield;
          getEfield (c.mid, c.efield);
          prec_t f = c.efield * c.ab_normal;
          c.q = 0.05 * f * c.rr / PCONST;
        }
    }
}

bool PField::equalize ()
{
  switch (method)
    {
      case POINTSYSTEM:
        return equalize_pointsystem();
        break;
      case CHARGEFLOW:
        return equalize_chargeflow();
        break;
      case LINESYSTEM:
        return equalize_linesystem();
        break;
      default:
        abort();
        break;
    }
  return false;
}

bool PField::equalize_chargeflow ()
{
  TaskManager tmgr;

  size_t cacheline_length = points.size() + (pc_q == 0.0 ? 0 : 1);

  size_t cs = sizeof (vect3d) * links.size() * cacheline_length;
  cache = (vect3d*)malloc (cs);
  if ( cache )
    {
      Ephi::debug (Ephi::DEBUG_INFO, "PField::equalize: allocated %lu bytes of memory for a cache\n", (unsigned long)cs);
      vect3d *p = cache;
      for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
        {
          vect3d& pos = it->mid;
          for ( points_t::iterator jt = points.begin(); jt != points.end(); jt++ )
            {
              vect3d r = pos - jt->pos;
              prec_t rr = r * r;
              prec_t m = rr * prec_t_sqrt(rr);
              *p++ = vect3d(r.x / m, r.y / m, r.z / m);
            }

          if ( pc_q != 0.0 )
            {
              vect3d r = pos - pc_pos;
              prec_t rr = r * r;
              prec_t m = rr * prec_t_sqrt(rr);
              *p++ = vect3d(r.x / m, r.y / m, r.z / m);
            }
        }
    }

  std::vector<task> tasks;
  tasks.resize (THREADS_CPUCOUNT);
  for ( size_t i = 0; i < THREADS_CPUCOUNT; i++ )
    {
      size_t pc = links.size() / THREADS_CPUCOUNT;
      tasks[i].pf = this;
      tasks[i].begin = links.begin();
      if ( cache )
          tasks[i].cache = cache + i * pc * cacheline_length;
      else
          tasks[i].cache = 0;
      std::advance (tasks[i].begin, i * pc);
      if ( i != THREADS_CPUCOUNT - 1 )
        {
          tasks[i].end = tasks[i].begin;
          std::advance (tasks[i].end, pc);
        }
      else
        {
          tasks[i].end = links.end();
        }
    }

  size_t n = 0;
  while (true)
    {
      for ( size_t i = 0; i < THREADS_CPUCOUNT; i++ )
          tmgr.addTask (&tasks[i]);
      while ( tmgr.run() );

      for ( feeds_t::iterator it = feeds.begin(); it != feeds.end(); it++ )
        {
          PPoint& p = *it->point;
          prec_t sum = 0, len = 0;
          for ( PPoint::links_t::iterator jt = p.links.begin(); jt != p.links.end(); jt++ )
            {
              sum += getPotential ((*jt)->mid);
              len += (*jt)->len;
            }
          sum /= p.links.size();
          len /= p.links.size() * 2;
          prec_t target = it->potential - sum;
          it->cp = sum;
          it->q = 0.1 * target * len / PCONST;
        }

      for ( feeds_t::iterator it = feeds.begin(); it != feeds.end(); it++ )
        {
          PPoint& p = *it->point;
          if ( !(n % PRINTMOD ) && 0 )
            {
              char buf[128];
              p.pos.sprint (buf);
              printf ("%s: potential = %f, added q = %g\n", buf, prec2double(it->cp), prec2double(it->q));
            }
          p.charge += it->q;
          added_charge += it->q;
        }
      
      for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
        {
          PLink& l = *it;
#if 0
          if ( !(n % PRINTMOD ) )
            {
              char buf[128];
              l.efield.sprint (buf);
              printf ("%3zu-%3zu: %g %s\n", l.a->id, l.b->id, l.q, buf);
            }
#endif
          l.a->charge -= l.q;
          l.b->charge += l.q;
          l.q = 0;
        }

      if ( !(n % PRINTMOD ) && 0 )
          print();
      n++;
      if ( n == 40001 )
        break;
      if ( !(n % PRINTMOD) )
          if ( getchar() == 'q' )
              break;
      Ephi::debug (Ephi::DEBUG_INFO, "PField::equalize_chargeflow: %u/%u\n", (unsigned int)n, 300);
      //if ( n == 600 )
      if ( n == 300 )
          break;
    }

  if ( cache )
    {
      free (cache);
      cache = 0;
    }
//  print();
  return true;
}

bool PField::equalize_pointsystem ()
{
  size_t n = points.size();
  LinearSystem ls(n);

  prec_t **rows = ls.getRows();
  for ( points_t::iterator jt = points.begin(); jt != points.end(); jt++ )
    {
      vect3d pos = jt->pot_pos;
      prec_t b = jt->potential / (SPEED_OF_LIGHT*SPEED_OF_LIGHT*MAGNETIC_CONSTANT);
      if ( pc_q != 0 )
        {
          prec_t ret = 0;
          prec_t rr = (pos - pc_pos).magnitude();
          if ( rr < pc_rr )
            {
              ret += pc_q / pc_radius;
              ret += 4.0 * PREC_PI * pc_rho * (pc_rr - rr) / 6.0;
            }
          else
            {
              ret += pc_q / prec_t_sqrt (rr);
            }
          b -= ret;
          //b -= pc_q / (pos - pc_pos).length();
        }
      prec_t *ptr = *rows;
      for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
        {
          *ptr = 1.0 / (pos - it->pos).length();
          ptr++;
        }
      *ptr = b;
      rows++;
    }

  if ( !ls.solveGBS() )
      return false;

  added_charge = 0;
  const prec_t *result = ls.getResults();
  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      it->charge = *result++;
      added_charge += it->charge;
    }

#if 0
  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
      printf ("Point id=%u, q=%g, pot=%g\n", (unsigned int)it->id, prec2double(it->charge), prec2double(getPotential(it->pot_pos)));
#endif
  return true;
}

bool PField::equalize_linesystem ()
{
  size_t n = points.size();
  LinearSystem ls(n);
  ls.clear();

  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      PPoint& point = *it;
      vect3d pos = point.pot_pos;

      prec_t *row = ls.getRow (point.id);

      for ( links_t::iterator jt = links.begin(); jt != links.end(); jt++ )
        {
          PLink& c = *jt;
          vect3d p0 = c.a->pos;
          vect3d p1 = c.b->pos;
          vect3d ln = c.ab_normal;
          prec_t llen = c.len;

          size_t p0i, p1i;
          vect3d r0, r1;
          prec_t a, b;

          vect3d tv = pos - p0;
          prec_t tm = ln * tv;
          if ( tm <= 0 )
            {
              p0i = c.a->id;
              p1i = c.b->id;
              r0 = tv;
              r1 = pos - p1;
              a = -tm;
              b = -(ln * r1);
            }
          else
            {
              p0i = c.b->id;
              p1i = c.a->id;
              r0 = pos - p1;
              r1 = tv;
              a = (ln * r0);
              b = tm;
            }
          prec_t r0len = r0.length();
          prec_t r1len = r1.length();

          prec_t logv = prec_t_log ((b + r1len) / (a + r0len));
          prec_t f1 = a / llen;
          prec_t f2 = (r1len - r0len) / llen;

          prec_t a0 = (1 + f1) * logv - f2;
          prec_t a1 = -f1 * logv + f2;

          row[p0i] += a0;
          row[p1i] += a1;

          prec_t bv = point.potential / (SPEED_OF_LIGHT*SPEED_OF_LIGHT*MAGNETIC_CONSTANT);
          if ( pc_q != 0 )
            {
              prec_t ret = 0;
              prec_t rr = (pos - pc_pos).magnitude();
              if ( rr < pc_rr )
                {
                  ret += pc_q / pc_radius;
                  ret += 4.0 * PREC_PI * pc_rho * (pc_rr - rr) / 6.0;
                }
              else
                {
                  ret += pc_q / prec_t_sqrt (rr);
                }
              bv -= ret;
            }
          row[n] = bv;
        }
    }

  if ( !ls.solveGBS() )
      return false;

  const prec_t *result = ls.getResults();
  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
      it->rho = *result++;

  for ( links_t::iterator lit = links.begin(); lit != links.end(); lit++ )
    {
      lit->rhodiff_by_llen = (lit->b->rho - lit->a->rho) / lit->len;
      added_charge += lit->len * lit->a->rho + 0.5 * lit->len * (lit->b->rho - lit->a->rho);
    }

  return true;
}

std::string PField::getPovrayScript ()
{
  char buf[1024];
  std::string ret;

  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      PPoint& p = *it;
      sprintf (buf, "sphere { <%f, %f, %f>, %f pigment { rgb <0.8, 1.0, 0.8> } }\n",
          prec2double (p.pos.x), prec2double (p.pos.y), prec2double (p.pos.z),
          prec2double (resolution / 6));
      ret.append (buf);

      if (0)
      if ( method == POINTSYSTEM || method == LINESYSTEM )
        {
          sprintf (buf, "sphere { <%f, %f, %f>, %f pigment { rgb <0.0, 0.8, 0.0> } }\n",
              prec2double (p.pot_pos.x), prec2double (p.pot_pos.y), prec2double (p.pot_pos.z),
              prec2double (resolution / 3));
          ret.append (buf);
        }
    }
  
  for ( links_t::iterator jt = links.begin(); jt != links.end(); jt++ )
    {
      PLink& l = *jt;
      PPoint& p1 = *l.a;
      PPoint& p2 = *l.b;
      sprintf (buf, "  cylinder { <%f, %f, %f>, <%f, %f, %f>, %f texture {StaticLineTexture}}\n",
          prec2double (p1.pos.x), prec2double (p1.pos.y), prec2double (p1.pos.z),
          prec2double (p2.pos.x), prec2double (p2.pos.y), prec2double (p2.pos.z),
          prec2double (resolution / 8));
      ret.append (buf);
    }

  return ret;
}

std::string PField::getChargeConcentrationScript ()
{
  prec_t nmin = -9e99;
  prec_t nmax = 9e99;
  prec_t pmin = 9e99;
  prec_t pmax = -9e99;

  std::string ret;

  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      prec_t q = it->charge;
      if ( method == LINESYSTEM )
          q = it->rho;
      //printf ("%u: q=%g\n", (unsigned int)it->id, q);
      if ( q < 0 )
        {
          if ( q > nmin )
              nmin = q;
          if ( q < nmax )
              nmax = q;
        }
      else
        {
          if ( q < pmin )
              pmin = q;
          if ( q > pmax )
              pmax = q;
        }
    }

  vect3d positive(1, 0, 0);
  vect3d neutral(0, 0.5, 0);
  vect3d negative(0, 0, 1);

  prec_t nm = nmin - nmax;
  prec_t pm = pmax - pmin;

  //printf ("nmin=%g, nmax=%g, nm=%g\n", nmin, nmax, nm);

  for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
    {
      if ( it->flags & PPoint::NO_CDIST )
          continue;
      prec_t q = it->charge;
      if ( method == LINESYSTEM )
          q = it->rho;
      vect3d c;
      if ( q < 0 )
        {
          c = (q / nmax) * negative + (1 - q/nmax) * neutral;
          c = ((nmin - q) / nm) * negative + (1 - (nmin - q) / nm) * neutral;
          //char buf[128];
          //printf ("%g: %s\n", q, c.sprint(buf));
        }
      else
        {
          c = (q / pmax) * positive + (1 - q / pmax) * neutral;
          c = ((q - pmin) / pm) * positive + (1 - (q - pmin) / pm) * neutral;
        }

      char buf[1024];
      sprintf (buf, "sphere { <%f, %f, %f>, %f pigment { rgb <%f, %f, %f> } }\n",
          prec2double (it->pos.x), prec2double (it->pos.y), prec2double (it->pos.z),
          prec2double (0.5 * resolution),
          prec2double (c.x), prec2double (c.y), prec2double (c.z));
      ret += buf;
    }

  return ret;
}

void PField::getEfield (const vect3d& pos, vect3d& efield)
{
  efield.clear();

  if ( method == LINESYSTEM )
    {
      for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
        {
          PLink& c = *it;
          vect3d p0 = c.a->pos;
          vect3d p1 = c.b->pos;
          vect3d ln = c.ab_normal;
          prec_t rho0 = c.a->rho;
          prec_t rho1 = c.b->rho;
          prec_t rhodiff_by_llen = c.rhodiff_by_llen;

          vect3d r0, r1;
          prec_t a, b;
          prec_t u, v;
          vect3d n;

          vect3d tv = pos - p0;
          prec_t tm = ln * tv;
          if ( tm <= 0 )
            {
              r0 = tv;
              r1 = pos - p1;
              a = -tm;
              b = -(ln * r1);
              u = rho0 - a * rhodiff_by_llen;
              v = rhodiff_by_llen;
              n = ln;
            }
          else
            {
              r0 = pos - p1;
              r1 = tv;
              a = (ln * r0);
              b = tm;
              u = rho1 - a * -rhodiff_by_llen;
              v = -rhodiff_by_llen;
              n = -ln;
            }
          prec_t r0len = r0.length();
          prec_t r1len = r1.length();
          prec_t af = a + r0len;
          prec_t bf = b + r1len;
          prec_t lf = log(bf / af);
          vect3d r0n = r0 / r0len;
          vect3d r1n = r1 / r1len;
          vect3d add = v * (lf * n + r1n - r0n) + u * ((r1n - n) / bf - (r0n - n) / af);
          efield -= add;
        }
    }
  else
    {
      // XXX this can be SSE-optimized
      for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
        {
          vect3d r = pos - it->pos;
          prec_t rr = r * r;
          efield += (it->charge / (rr * prec_t_sqrt(rr))) * r;
        }
    }

  if ( pc_q != 0 )
    {
      vect3d r = pos - pc_pos;
      prec_t rr = r.magnitude();
      if ( rr < pc_rr )
        {
          efield += (pc_q / (pc_rr * pc_radius)) * r;
        }
      else
        {
          efield += (pc_q / (rr * prec_t_sqrt(rr))) * r;
        }
    }

  efield *= PCONST;
}

prec_t PField::getPotential (const vect3d& pos)
{
  prec_t ret = 0;

  if ( method == LINESYSTEM )
    {
      for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
        {
          PLink& c = *it;
          vect3d p0 = c.a->pos;
          vect3d p1 = c.b->pos;
          vect3d ln = c.ab_normal;
          prec_t rho0 = c.a->rho;
          prec_t rho1 = c.b->rho;
          prec_t rhodiff_by_llen = c.rhodiff_by_llen;

          vect3d r0, r1;
          prec_t a, b;
          prec_t u, v;
          vect3d n;

          vect3d tv = pos - p0;
          prec_t tm = ln * tv;
          if ( tm <= 0 )
            {
              r0 = tv;
              r1 = pos - p1;
              a = -tm;
              b = -(ln * r1);
              u = rho0 - a * rhodiff_by_llen;
              v = rhodiff_by_llen;
            }
          else
            {
              r0 = pos - p1;
              r1 = tv;
              a = (ln * r0);
              b = tm;
              u = rho1 - a * -rhodiff_by_llen;
              v = -rhodiff_by_llen;
            }
          prec_t r0len = r0.length();
          prec_t r1len = r1.length();
          prec_t add = (u * prec_t_log ((b + r1len) / (a + r0len)) + v * (r1len - r0len));
          ret += add;
        }
    }
  else
    {
      // XXX this can be SSE-optimized
      for ( points_t::iterator it = points.begin(); it != points.end(); it++ )
          ret += it->charge / (pos - it->pos).length();
    }

  if ( pc_q != 0 )
    {
      prec_t rr = (pos - pc_pos).magnitude();
      if ( rr < pc_rr )
        {
          ret += pc_q / pc_radius;
          ret += 4.0 * PREC_PI * pc_rho * (pc_rr - rr) / 6.0;
        }
      else
        {
          ret += pc_q / prec_t_sqrt (rr);
        }
    }

  ret *= PCONST;
  return ret;
}

PLink& PField::connect (PPoint& a, PPoint& b)
{
#if 1
  for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
      assert ( !(it->a == &a && it->b == &b || it->a == &b && it->b == &a) );
  links.push_back (PLink (&a, &b));
  a.links.push_back (&links.back());
  b.links.push_back (&links.back());
  return links.back();
#else
  PPoint& m = addPoint (0.5 * (a.pos + b.pos));
  links.push_back (PLink (&a, &m));
  a.links.push_back (&links.back());
  m.links.push_back (&links.back());
  links.push_back (PLink (&m, &b));
  m.links.push_back (&links.back());
  b.links.push_back (&links.back());
#endif
}

PLink* PField::safe_connect (PPoint& a, PPoint& b)
{
  if ( &a == &b ) return 0;
  for ( links_t::iterator it = links.begin(); it != links.end(); it++ )
      if ( it->a == &a && it->b == &b || it->a == &b && it->b == &a )
          return 0;
  links.push_back (PLink (&a, &b));
  a.links.push_back (&links.back());
  b.links.push_back (&links.back());
  return &links.back();
}

PLink& PField::split (PLink& link, PPoint& mid)
{
  PPoint& a = *link.a;
  PPoint& b = *link.b;
  link = PLink (&a, &mid);
  mid.links.push_back (&link);
  b.links.remove (&link);
  return connect (mid, b);
}

PPoint::PPoint (const vect3d& pos, const vect3d& pot_pos, prec_t potential) :
    pos(pos), pot_pos(pot_pos), potential(potential), charge(0), rho(0), inc(0)
{
  cur_efield.clear();
  flags = 0;
}

void StaticPotentialField::addFields (const vect3d& pos, vect3d& bfield, vect3d& efield)
{
  vect3d ef;
  pfield.getEfield (pos, ef);
  efield += ef;
}

void StaticPotentialField::addPotential (const vect3d& pos, prec_t& potential)
{
  potential += pfield.getPotential (pos);
}

void StaticPotentialField::draw(Screen& screen, const Screen::color& c)
{
}

std::string StaticPotentialField::get_povray_script ()
{
  return pfield.getPovrayScript();
}

bool StaticPotentialField::inContact (const vect3d& pos)
{
  prec_t mag = (pfield.resolution / 2 ) * (pfield.resolution / 2);
  prec_t mag2 = (pfield.resolution / 4 ) * (pfield.resolution / 4);

  if ( pfield.method == PField::LINESYSTEM )
    {
      PField::links_t& links = pfield.links;
      for ( PField::links_t::iterator it = links.begin(); it != links.end(); it++ )
        {
          PLink& c = *it;
          vect3d r0 = pos - c.a->pos;
          prec_t r0r0 = r0 * r0;
          prec_t ld = c.ab_normal * r0;
          if ( r0r0 - ld * ld > mag2  )
              continue;
          if ( ld >= 0 && ld <= c.len )
              return true;
          if ( r0r0 <= mag2 || (c.b->pos - pos).magnitude() <= mag2 )
              return true;
        }
      return false;
    }

  for ( PField::points_t::iterator it = pfield.points.begin(); it != pfield.points.end(); it++ )
    {
      if ( (it->get_pos() - pos).magnitude() <= mag )
          return true;
    }

  if ( pfield.pc_q != 0 && pfield.pc_radius == 0 )
    {
      prec_t mag2 = (pfield.resolution / 2) * (pfield.resolution / 2);
      mag2 = 0.05 * 0.05;
      if ( (pos - pfield.pc_pos).magnitude() <= mag2 )
          return true;
    }
  return false;
}


