//
// Ephi - simulation of magnetic fields and particles
// Copyright (C) 2007 Indrek Mandre <indrek(at)mare.ee>
// For more information please see http://www.mare.ee/indrek/
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#ifndef __potential_hpp__
#define __potential_hpp__

#include <list>
#include <string>
#include <vector>

class PField;
class PLink;

class PPoint
{
public:
  PPoint (const vect3d& pos, const vect3d& ppos = vect3d(0, 0, 0), prec_t potential = 0);

  const vect3d& get_pos () const { return pos; }

private:
  typedef std::list<PLink*> links_t;
  links_t links;

  size_t id;
  vect3d pos, pot_pos;
  prec_t potential;
  prec_t charge;
  prec_t rho;
  enum eflags { NO_CDIST = 1 };
  size_t flags;

  vect3d cur_efield;
  prec_t inc;

  void step() { charge += inc; inc = 0; }

  friend class PField;
  friend class PLink;
  friend class StaticPotentialField;
};

class PLink
{
public:
  PLink(PPoint *a, PPoint *b) : a(a), b(b), q(0)
  {
    ab = b->pos - a->pos;
    ab_normal = ab.normal();
    rr = ab*ab;
    len = prec_t_sqrt (rr);
    mid = 0.5 * (a->pos + b->pos);
    cc_normal = (0.05 * rr) * ab_normal;
    abr = (1 / (ab * ab)) * ab;
    rhodiff_by_llen = (b->rho - a->rho) / len;
  }

private:
  PPoint *a, *b;
  vect3d ab;
  vect3d ab_normal;
  vect3d abr;
  vect3d cc_normal;
  prec_t rr;
  prec_t len;
  vect3d mid;
  prec_t rhodiff_by_llen;

  prec_t q;
  vect3d efield;
  PPoint *tmp;

  friend class PField;
  friend class PPoint;
  friend class StaticPotentialField;
};

class StaticPotentialField;

class PField
{
public:
  enum emethod_t { POINTSYSTEM = 1, CHARGEFLOW, LINESYSTEM };

  #define UNKNOWN_POTENTIAL 9e99

  PField (emethod_t method = POINTSYSTEM, prec_t resolution = 0.02);
  ~PField ();

  PPoint& addPoint (const vect3d& pos, const vect3d& ppos = vect3d(0, 0, 0), prec_t potential = 0);
  void addFeedPoint (PPoint& p, prec_t potential);

  void addSphere (const vect3d& p, prec_t radius, prec_t potential = 0, bool backawrds = false);
  void addTorus (const vect3d& p, const vect3d& normal, prec_t radius, prec_t wr, prec_t potential = 0);
  void addStrip (const vect3d& p0, const vect3d& p1, size_t n_segments = 0);
  void addT (const vect3d& p0, const vect3d& p1, const vect3d& p2, const vect3d& p3);

  bool equalize();

  std::string getPovrayScript ();
  std::string getChargeConcentrationScript ();

  void getEfield(const vect3d& pos, vect3d& efield);
  prec_t getPotential(const vect3d& pos);

  void setPointCharge (const vect3d& pos, prec_t q);
  void setUniformCharge (const vect3d& pos, prec_t q, prec_t radius);

  PLink& connect (PPoint& a, PPoint& b);
  PLink* safe_connect (PPoint& a, PPoint& b);
  PLink& split (PLink& link, PPoint& mid);

  void print();

  prec_t getAddedCharge () const { return added_charge; }
  prec_t getAveragePotential();
  void setPotential (prec_t pot);

  void save (const char *path);
  void save (const std::string& path) { save (path.c_str()); }
  bool load (const char *path);
  bool load (const std::string& path) { return load (path.c_str()); }

  void set_no_cdist (bool ncd) { no_cdist = ncd; }

  void getStatistics (size_t& pointcount, size_t& linkcount) { pointcount = points.size(); linkcount = links.size(); }

  prec_t getResolution() const { return resolution; }
  prec_t setResolution(prec_t res) { prec_t ores = resolution; resolution = res; return ores; }

private:
  emethod_t method;
  prec_t resolution;
  size_t pcounter;
  bool no_cdist;

  typedef std::list<PPoint> points_t;
  points_t points;

  struct fpoint_t
  {
    fpoint_t() : cp(0), q(0) { }
    PPoint *point;
    prec_t potential;
    prec_t cp;
    prec_t q;
  };

  typedef std::list<fpoint_t> feeds_t;
  feeds_t feeds;

  prec_t added_charge;

  prec_t pc_q;
  vect3d pc_pos;
  prec_t pc_radius;
  prec_t pc_rr;
  prec_t pc_rho;

  typedef std::list<PLink> links_t;
  links_t links;

  struct task : Task
  {
    PField *pf;
    links_t::iterator begin;
    links_t::iterator end;
    vect3d *cache;
    void execute () { pf->calc_efield (begin, end, cache); }
  };

  void calc_efield (const links_t::iterator& begin, const links_t::iterator& end, const vect3d* cache);
  bool equalize_chargeflow();
  bool equalize_pointsystem();
  bool equalize_linesystem();

  vect3d * cache;

  friend class StaticPotentialField;
};

class StaticPotentialField : public StaticElement
{
public:
  StaticPotentialField (PField &pfield) : pfield(pfield) { }

  void addFields (const vect3d& pos, vect3d& bfield, vect3d& efield);
  void addPotential (const vect3d& pos, prec_t& potential);
  void draw(Screen& screen, const Screen::color& c = Screen::BLACK);
  std::string get_povray_script ();
  bool inContact (const vect3d& pos);

private:
  PField& pfield;
};

#endif // !__potential_hpp__

